'''
  this is a visualization class
'''
import statsmodels.tsa.api as tsa
import pandas as pd
import zerorpc


class app():
    '''Access to the services provided by the app
    Args:
        appParams (dict) : parameters needed to connect to app
    '''
    
    def __init__(self,port=None,modelHub=None):
        self.port = port 
        self.portStub            = 'tcp://127.0.0.1:'
        self.portGuess           = 4242 #check index.js in app for this number
        self.portSearchIncrement = 123  #check index.js in app for this number
        self.client = zerorpc.Client()
        self.modelHub = modelHub
        self.modelResult = {}  #result of a fitted model
        #connect:
        self.connect()
        
    def connect(self):
        '''
          connects to the pypungi
        '''
        if self.port is not None:
            self.client.connect(self.port)
        else:
            keepTrying = 1
            while keepTrying == 1:
                try: 
                    self.client.connect(self.portStub + str(self.portGuess))
                    self.ping()
                    self.port = self.portGuess
                    keepTrying = 0
                except:  
                    self.portGuess = self.portGuess + self.portSearchIncrement
    
    def ping(self,text='ping'):
        return(self.client.ping(text))
    
    def plot(self,data,select=None):
        '''
          given data and plottype process and send it to app
        '''
        
        if select is None:
            #if isinstance(data,pd.DataFrame) or isinstance(data,pd.Series):
            #    select = 'tseries'
            #else:
            select = 'line'
        
        rawdata = self._transform(data,select)
        #self.client.plot({'data':transmitData, 'plottype':plottype})
        self.client.plot({'rawdata':rawdata,'select':select})
    
    def model(self,modelName,*args,**kwargs):
        '''send model id (eg ols) and the 'app' part of the fitted model'''
        #self.model = statsModel(modelName,*args,**kwargs)
        #self.client.model(  {'id':modelName,'table':fitmodel.model['app'] } )
        self.modelResult = self.modelHub(modelName,*args,**kwargs)
        self.client.model(self.modelResult['app'])
     
    def _transform(self,data,select):
        
        if select in ['line','scatter']:
            return(self._transformDataTypeToDictionary(data)) #if series, transform to dataframe
        elif select in ['tseries']:
            sendData = {}
            sendData['rawdata'] = self._transformDataTypeToDictionary(data)
            sendData['tables'] = self._tsaStats(sendData['rawdata'])
            return(sendData)
        elif select in ['table']:
            return(self._transformDataTypeToDictionary(data,format='split'))
        else:
            return(self._transformDataTypeToDictionary(data))
    
    def _transformDataTypeToDictionary(self,data,format='list'):
        '''
          given data of certain type, transform it to the most common used dictionary.
          if a type of plot, as indicated by select, needs a specific data transformation, do it there.
          Else, introduce a new input var (select) here and use it. 
          
          for data:
          format should be list: eg {'a':[1,2,3,4], 'b':[1,2,3,4]}
          
          for tables:
          default format should be split: eg {'index': [0, 1], 'columns': [0, 1, 2], 'data': [[1, 2, 3], [1, 2, 3]]}
        '''
        if isinstance(data,pd.DataFrame) or isinstance(data,pd.Series):
            if 'date' in data.columns or 'index' in data.columns:
                return(pd.DataFrame(data).rename({'data':'index'},axis='columns').astype({'index':str}).to_dict(format))
            else:
                return(pd.DataFrame(data).assign(index = data.index).astype({'index':str}).to_dict(format))
        elif isinstance(data,list):
                return({'index':list(range(len(data))), '':data    })
        else:
            return(data)
    
    def _tsaStats(self,rawdata):
        '''for data in 'records' format: {'a':[1,2,3]}, calculate time series stats'''
        #TODO: it would be better to round on the app side and give user the option to select precission
        tableStats = {}
        
        #STATIONARITY TESTS
        kpssTable = {'columns': ['variable','stat','p-value','lags','10%','5%', '2.5%', '1%'] }
        data = []
        for var in rawdata.keys():
            if var != 'index':
                kpssTest = list(tsa.kpss(rawdata[var],nlags='auto'))
                kpssTest = [var] + [float(round(x,4)) for x in kpssTest[0:3] + list(kpssTest[3].values())] #float bc can't serialize numpy.int64
                data += [kpssTest]
        kpssTable['data'] = data
        tableStats['kpss'] = kpssTable
        
        adfTable = {'columns': ['variable','stat','p-value','lags','N obs', '1%','5%', '10%', 'IC'] }
        data = []
        for var in rawdata.keys():
            if var != 'index':
                adfTest = list(tsa.adfuller(rawdata[var]))
                adfTest = [var] + [float(round(x,2)) for x in adfTest[0:4] + list(adfTest[4].values())] + [round(adfTest[5],1)] #float bc can't serialize numpy.int64
                data += [adfTest]
        adfTable['data'] = data
        tableStats['adf'] = adfTable
        
        tableStats['describe'] = pd.DataFrame(rawdata).describe().T.reset_index().rename({'index':'variable'}, axis=1).to_dict('split')
        return(tableStats)
        
    
    def stash(self,data):
        
        if isinstance(data,pd.DataFrame):
            rawdata = data.to_dict('list')
        elif isinstance(data,pd.Series):
            rawdata = data.to_dict()
        else:
            rawdata = data
        
        content = {'data':rawdata, 'type': str(type(data))}
        
        self.client.stash(content)
    
    def getStash(self):
        stash = self.client.getStash('')
        if stash['type'] == "<class 'pandas.core.frame.DataFrame'>":
            return(pd.DataFrame(stash['data']))
        elif stash['type'] == "<class 'pandas.core.series.Series'>":
            return(pd.Series(stash['data']))
        else:
            return(stash)
    
    def download(self):
        download = self.client.load('')
        return(download)


'''
  This is a "transformation class" takes data in a format and returns data in a format...
'''

import statsmodels.api as sm
import pandas as pd

class ModelHub:
    #distributes model to the right model class 
    def __init__(self):
        self.model = {}
        self.statsModels = statsModels()
    def __call__(self,modelName,*args,**kwargs):
        if modelName in ['ols']:
            self.model = self.statsModels(modelName,*args,**kwargs)
            return(self.model)
        else :
            outlets = ['model','fit','app','latex','db','print']
            self.model = { outlet : dict(id = modelName,args = args,kwargs = kwargs) for outlet in outlets }
            return(self.model)


class statsModels():
    '''
      self.model should have: "app" key - the data to be sent to app
                              "latex" key - data to be sent to a latex doc
                              "db"  key - data to be stored in a db
    
    '''
    def __init__(self):
        pass
    
    def __call__(self,modelName,*args,**kwargs):
        self.modelName = modelName
        self.args = args
        self.kwargs = kwargs
        self.model = {}
    
        if self.modelName == 'ols':
            self.model['model'] = sm.OLS(*self.args,**self.kwargs)
            self.model['fit']   = self.model['model'].fit()
            self.model['app']   = dict(
                id='ols',
                table = self.model['fit'].summary2().as_html(), 
                residuals = pd.DataFrame({
                    'resid':list(self.model['fit'].resid)}).reset_index().to_dict('list'),
                predict = pd.DataFrame({ 
                    'value' :  list(self.model['fit'].model.endog), 
                    'prediction' : list(self.model['fit'].predict())}
                    ).reset_index().to_dict('list')
            )
            self.model['latex'] = self.model['fit'].summary2().as_latex()
            self.model['db']    = ""
            self.model['print'] = self.model['fit'].summary2()
    
        return(self.model)


if __name__ == '__main__':
    import numpy as np
    e = np.random.normal(0,1,[100,1])
    x = np.arange(100).reshape([100,1])
    y = 3*x + e
    model = ModelHub()
    model('ols',y,100*x)
    


'''
'''
import pandas as pd

import statsmodels.tsa.api as tsa
from pypungi.models import ModelHub 
from pypungi.outlets import Outlets  


class hub():
    '''
       
       Default visualization is app. Other outlets are called by first 
       indicating that it will be used.
       Connects
       suggest run:
       
       import pungipy 
       
       pp = pungipy.link() 
       pp.hello()
       pp.plot('gdp')
    '''
    def __init__(self,outlets=['app'],port=None,dbInfo=None,latexPath=None):
        #visualization classes
        self.modelHub = ModelHub()  #transforms data before outlet sends it.  call method evaluates the model: modelName, *args, **kwargs
        self.outlets = Outlets(outlets,port=None,dbInfo=None,latexPath=None,model=self.modelHub)  
        self.app   = self.outlets.outlets['app']['link']
        self.db    = self.outlets.outlets['db']['link']
        self.latex = self.outlets.outlets['latex']['link']
        self.print = self.outlets.outlets['print']['link']
        #transformation classes
    
    
    def __call__(self):
        return(self.app)

if __name__ == '__main__':
    #v = link()  #notice, need a pungi session running for this to work.
    #v.checkConnection()
    #v.plot('gdp')  
    
    #pp = link()
    #pp.plot([1,2,3,4])
    #pp.plot([1,2,3,4,1,1,2,3,1,3,44,1,1],'tseries')
    
    #import numpy as np 
    # 
    #m = pd.DataFrame(np.random.normal(0,1,(100,4)), columns = ['a','asdasdf asdfas','c','d'])
    
    #pp.plot(m,'tseries')
    
    #import numpy
    #import statsmodels.api as sm
    #from main import link
    
    #X = sm.datasets.spector.load(as_pandas=False)
    #Y = sm.add_constant(X.exog, prepend=False)
    #pp = link()
    #pp.model('ols',X.endog,Y)
    
    #pp = hub(['app','print'])
    pp = hub()()
    
    
    pp.plot([1,2,3,4,5],'scatter')
    import numpy as np
    e = np.random.normal(0,1,[100,1])
    x = np.arange(100).reshape([100,1])
    y = 3*x + e
    pp.model('ols',x,3*y)    
    #pp.model('policybrief','hi')


#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pypungi']

package_data = \
{'': ['*']}

install_requires = \
['statsmodels', 'zerorpc']

setup(name='pypungi',
      version='0.11.15',
      description='Helper tool',
      author='James Otterson',
      author_email='jjotterson@gmail.com',
      url='https://github.com/jjotterson',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )



'''
  Link class - the links take data and send them to outlets.  
'''
import statsmodels.tsa.api as tsa
import pandas as pd
import zerorpc

from pypungi.app import app

class Outlets:
    '''Mediates connection to outlet
    
    Attributes:
        connectTo (list) : which outlets to use - app, console, latex, db
        db (dict)        : data needed to connect to databases
        latex (dict)     : data needed to connect to a folder of latex outputs as plots tables
        console (dict)   : data needed to print to console 
    '''
    def __init__(self,outlets=['app'],port=None,dbInfo=None,latexPath=None,model=None):
        '''
        Args:
            connectTo (list) : which outlets - app, console, latex, db 
            port (str)       : the tcp of the app
            db (str)         : db connection info
            latexPath (str)  : path where to save latex outputs
            model (obj)      : a callable object (has __call__) that takes data and transforms it for each outlet
        '''
        self.selectedOutlets = outlets
        self.outlets = dict(
            app  = {
                'port':port,
                'connected':'app' in outlets,
                'link': self.linkApp(port,'app' in outlets,model)
            },
            db = {
                'dbInfo' : dbInfo,
                'connected':'db' in outlets,
                'link': self.linkDB(dbInfo,'db' in outlets,model)
            },
            latex = {
                'latexPath' : latexPath,
                'connected':'latex' in self.selectedOutlets,
                'link': self.linkLatex(latexPath,'latex' in outlets,model)                
            },
            print = {
                'connected':'print' in self.selectedOutlets,
                'link': self.linkPrint('print' in outlets,model)                
            },
        )   
    
    def notConnected(self,*args,**kwargs):
        print('outlet not connected')
    
    def linkApp(self,port,start,model):
        '''connects to app
        
        '''
        if start :
            return(app(port,model))
        else :
            return(self.notConnected)
    
    def linkDB(self,dbInfo,start,model):
        if start :
            print('db connected')
        else :
            return(self.notConnected)        
        
    def linkLatex(self,latexPath,start,model):
        if start :
            print('latex connected')
        else :
            return(self.notConnected)        
            
    def linkPrint(self,start,model):
        if start :
            print('print connected')
        else :
            return(self.notConnected)        
    
    def __call__(self,start):
        if start :
            print('connected')
        else :
            return(self.notConnected)        



'''Helper tool
   import pypungi 
   pp = pypungi.app() #or pass tcp/ip
   or
   pp = pypungi.hub()()

'''

__version__ = "0.11.15"

from pypungi.main import hub 
from pypungi.app import app



if __name__ == '__main__':
    pass
    #import pypungi
    #
    #pp = pypungi.link()
    #pp.stash('hi')
    #pp.getStash()
    